import React, {useCallback, useContext, useMemo, useState} from "react";
import {v4 as uuidv4} from "uuid";

/**
 * Create Kernel context.
 * @param defaultValue Kernel context default value.
 */
export function createKernelContext<T>(defaultValue: T): KernelContext<T>
{
	return {
		uuid: uuidv4(),
		defaultValue: defaultValue,
	};
}

/**
 * Kernel context definition object.
 */
export interface KernelContext<T>
{
	/**
	 * Kernel context UUID.
	 */
	uuid: string;

	/**
	 * Kernel context default value.
	 */
	defaultValue: T;
}

/**
 * Kernel context setter function type.
 */
export type KernelContextSetter<IdentifierType extends string|symbol|number = string, ValueType = any> = (identifier: IdentifierType, value: ValueType) => void;

/**
 * Kernel contexts type.
 */
export type KernelContexts = Record<string, any>;

/**
 * Kernel context dispatcher function type.
 */
export type KernelContextDispatcher<ValueType = any> = (value: ValueType) => void;

/**
 * Kernel global context data.
 */
export interface KernelGlobalContextData
{
	contexts: KernelContexts;
	setContext: KernelContextSetter;
}

/**
 * React Kernel global context.
 */
export const KernelGlobalContext = React.createContext<KernelGlobalContextData>({
	contexts: {},
	setContext: () => {},
});

/**
 * Kernel global context provider.
 */
export function KernelGlobalContextProvider({children}: React.PropsWithChildren<{}>)
{
	// Kernel contexts initialization.
	const [kernelContexts, setKernelContexts] = useState<KernelContexts>({});

	/**
	 * Change kernel contexts.
	 */
	const changeKernelContexts = useCallback((kernelContextsUpdate: Partial<KernelContexts>) => {
		setKernelContexts({
			...kernelContexts,
			...kernelContextsUpdate,
		});
	}, [kernelContexts, setKernelContexts]);

	/**
	 * Kernel context setter function.
	 */
	const kernelContextSetter = useCallback((identifier: string, data: any) => (
		changeKernelContexts({
			[identifier]: data,
		})
	), [changeKernelContexts]);

	// Initialize kernel global context value.
	const kernelContextValue = useMemo<KernelGlobalContextData>(() => ({
		contexts: kernelContexts,
		setContext: kernelContextSetter,
	}), [kernelContexts, kernelContextSetter]);

	return (
		<KernelGlobalContext.Provider value={kernelContextValue}>
			{children}
		</KernelGlobalContext.Provider>
	);
}

/**
 * Get kernel global context data.
 */
function useKernelGlobalContext(): KernelGlobalContextData
{
	return useContext(KernelGlobalContext);
}

/**
 * Initialize or get kernel context with given identifier.
 * @param context Context object.
 */
export function useKernelContext<T>(context: KernelContext<T>): [T, KernelContextDispatcher<T>]
{
	/**
	 * Get kernel global context.
	 */
	const kernelGlobalContext = useKernelGlobalContext();

	/**
	 * Function to set kernel context.
	 */
	const setContext = useCallback(
		(newValue: T) => kernelGlobalContext.setContext(context.uuid, newValue),
		[kernelGlobalContext.setContext],
	);

	// Return kernel context and its dispatcher.
	return [kernelGlobalContext.contexts?.[context.uuid] ?? context.defaultValue, setContext];
}
